@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,6 +24,8 @@
 # are None, at least to the point of making it through __init__. This is
 # so that prediction of images works.
 
+from __future__ import print_function
+
 import renpy.display
 from renpy.display.render import render
 
@@ -315,9 +317,7 @@ class Dissolve(Transition):
         The time the dissolve will take.
 
     `alpha`
-        If true, the dissolve will alpha-composite the result of the transition
-        with the screen. If false, the result of the transition will replace the
-        screen, which is more efficient.
+        Ignored.
 
     `time_warp`
         A function that adjusts the timeline. If not None, this should be a
@@ -363,10 +363,10 @@ class Dissolve(Transition):
         width = min(top.width, bottom.width)
         height = min(top.height, bottom.height)
 
-        rv = renpy.display.render.Render(width, height, opaque=not self.alpha)
+        rv = renpy.display.render.Render(width, height, opaque=not (self.alpha or renpy.config.dissolve_force_alpha))
 
         rv.operation = renpy.display.render.DISSOLVE
-        rv.operation_alpha = self.alpha
+        rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = complete
 
         rv.blit(bottom, (0, 0), focus=False, main=False)
@@ -402,12 +402,10 @@ class ImageDissolve(Transition):
         completed one step of dissolving in.
 
     `reverse`
-        If true, black pixels will dissolve in before white pixels.
+        If True, black pixels will dissolve in before white pixels.
 
     `alpha`
-        If true, the dissolve will alpha-composite the result of the transition
-        with the screen. If false, the result of the transition will replace the
-        screen, which is more efficient.
+        Ignored.
 
     `time_warp`
         A function that adjusts the timeline. If not None, this should be a
@@ -499,7 +497,7 @@ class ImageDissolve(Transition):
         width = min(bottom.width, top.width, image.width)
         height = min(bottom.height, top.height, image.height)
 
-        rv = renpy.display.render.Render(width, height, opaque=not self.alpha)
+        rv = renpy.display.render.Render(width, height, opaque=not (self.alpha or renpy.config.dissolve_force_alpha))
 
         complete = st / self.delay
 
@@ -507,7 +505,7 @@ class ImageDissolve(Transition):
             complete = self.time_warp(complete)
 
         rv.operation = renpy.display.render.IMAGEDISSOLVE
-        rv.operation_alpha = self.alpha
+        rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = complete
         rv.operation_parameter = self.ramplen
 
@@ -537,8 +535,7 @@ class AlphaDissolve(Transition):
         The time the transition takes, before ending.
 
     `alpha`
-        If true, the image is composited with what's behind it. If false,
-        the default, the image is opaque and overwrites what's behind it.
+        Ignored.
 
     `reverse`
         If true, the alpha channel is reversed. Opaque areas are taken
@@ -590,7 +587,7 @@ class AlphaDissolve(Transition):
         rv = renpy.display.render.Render(width, height, opaque=not self.alpha)
 
         rv.operation = renpy.display.render.IMAGEDISSOLVE
-        rv.operation_alpha = self.alpha
+        rv.operation_alpha = self.alpha or renpy.config.dissolve_force_alpha
         rv.operation_complete = 256.0 / (256.0 + 256.0)
         rv.operation_parameter = 256
 
@@ -614,7 +611,7 @@ class CropMove(Transition):
 
     Returns a transition that works by cropping a scene and positioning it on the
     screen. This can be used to implement a variety of effects, all of which
-    involved changing rectangular slices of scenes.
+    involve changing rectangular slices of scenes.
 
     `time`
         The time the transition takes.
@@ -873,7 +870,7 @@ class PushMove(Transition):
     :name: PushMove
 
     Returns a transition that works by taking the new scene and using it to
-    "push" the old scene off the screen. 
+    "push" the old scene off the screen.
 
     `time`
         The time the transition takes.
