@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,10 +19,13 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
+import collections
 import renpy.sl2
 import renpy.sl2.slast as slast
 
-# A list of style prefixes that we know of.
+# A tuple of style prefixes that we know of.
 STYLE_PREFIXES = [
     '',
     'insensitive_',
@@ -61,8 +64,8 @@ class Positional(object):
             parser.add(self)
 
 
-# Used to generate the documentation
-all_keyword_names = set()
+# This is a map from (prefix, use_style_prefixes) to a set of property names.
+properties = collections.defaultdict(set)
 
 
 class Keyword(object):
@@ -73,7 +76,7 @@ class Keyword(object):
     def __init__(self, name):
         self.name = name
 
-        all_keyword_names.add(self.name)
+        properties['', False].add(name)
 
         if parser:
             parser.add(self)
@@ -87,8 +90,7 @@ class Style(object):
     def __init__(self, name):
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(j + self.name)
+        properties['', True].add(self.name)
 
         if parser:
             parser.add(self)
@@ -103,8 +105,7 @@ class PrefixStyle(object):
         self.prefix = prefix
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(prefix + j + self.name)
+        properties[prefix, True].add(self.name)
 
         if parser:
             parser.add(self)
@@ -117,7 +118,7 @@ class Parser(object):
     # inside something that takes a single child.
     nchildren = "many"
 
-    def __init__(self, name):
+    def __init__(self, name, statement=True):
 
         # The name of this object.
         self.name = name
@@ -128,7 +129,8 @@ class Parser(object):
         self.keyword = { }
         self.children = { }
 
-        all_statements.append(self)
+        if statement:
+            all_statements.append(self)
 
         global parser
         parser = self
@@ -717,6 +719,11 @@ class ForParser(Parser):
         else:
             code = None
 
+        if l.match('index'):
+            index_expression = l.require(l.say_expression)
+        else:
+            index_expression = None
+
         l.require('in')
 
         expression = l.require(l.python_expression)
@@ -724,7 +731,7 @@ class ForParser(Parser):
         l.require(':')
         l.expect_eol()
 
-        rv = slast.SLFor(loc, name, expression)
+        rv = slast.SLFor(loc, name, expression, index_expression)
 
         if code:
             rv.children.append(slast.SLPython(loc, code))
@@ -742,7 +749,7 @@ class OneLinePythonParser(Parser):
     def parse(self, loc, l, parent):
 
         loc = l.get_location()
-        source = l.require(l.rest)
+        source = l.require(l.rest_statement)
 
         l.expect_eol()
         l.expect_noblock("one-line python")
@@ -938,7 +945,7 @@ class CustomParser(Parser):
 class ScreenParser(Parser):
 
     def __init__(self):
-        super(ScreenParser, self).__init__("screen")
+        super(ScreenParser, self).__init__("screen", statement=False)
 
     def parse(self, loc, l, parent, name="_name"):
 
